/************************************************************************/
/*  Treatment of ispell affix files.					*/
/*  Syntax is simplified. (No line breaks allowed in rule.)		*/
/************************************************************************/

#   include	<stddef.h>
#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<string.h>
#   include	<ctype.h>
#   include	<charnames.h>
#   include	<debugoff.h>
#   include	"affix.h"

/************************************************************************/
/* Applicability mask; macro's to use it.				*/
/************************************************************************/
typedef struct ApplicabilityMask
    {
    unsigned char	amBits[32];	/*  (256+7)/8	*/
    } ApplicabilityMask;

# define	amEmptyMask(a)	memset( (a)->amBits, 0x00, sizeof(a)->amBits )
# define	amFillMask(a)	memset( (a)->amBits, 0xff, sizeof(a)->amBits )
# define	amSet(a,v)	(a)->amBits[(v)/8] |= 1<<((v)%8)
# define	amClear(a,v)	(a)->amBits[(v)/8] &= ~(1<<((v)%8))
# define	amIsSet(a,v)	(((a)->amBits[(v)/8]&(1<<((v)%8)))!=0)

/************************************************************************/
/*  An affix rule, For simplicity a fixed maximum size for the strings	*/
/*  is assumed.								*/
/************************************************************************/
# define	AFFamMAXLENGTH	5
# define	AFFfrMAXLENGTH	5
# define	AFFtoMAXLENGTH	7

typedef struct AffixRule
    {
    int			arWhatSide;
    int			arShift;
    int			arMaskLength;
    int			arFromLength;
    int			arToLength;
    struct AffixRule *	arNext;
    ApplicabilityMask	arMasks[AFFamMAXLENGTH];
    unsigned char	arToReplace[AFFfrMAXLENGTH];
    unsigned char	arReplacement[AFFtoMAXLENGTH];
    } AffixRule;

# define BUL 300
static unsigned char buf[BUL+1];
static unsigned char word1[200+1];
static unsigned char word2[200+1];

static AffixRule *	AFFRules[256];

/************************************************************************/
/*  Read one affix character.						*/
/************************************************************************/
static int	affixChar(	const unsigned char *	buf,
				int *			pPos	)
    {
    int		pos= *pPos;

    switch( buf[pos] )
	{
	case 'A':
	    if  ( buf[pos+1] == '"' )
		{ (*pPos) += 2; return ISO_adieresis;		}
	    else{ (*pPos)++; return ISO_tolower( buf[pos] );	}

	case 'O':
	    if  ( buf[pos+1] == '"' )
		{ (*pPos) += 2; return ISO_odieresis;		}
	    else{ (*pPos)++; return ISO_tolower( buf[pos] );	}

	case 'U':
	    if  ( buf[pos+1] == '"' )
		{ (*pPos) += 2; return ISO_udieresis;		}
	    else{ (*pPos)++; return ISO_tolower( buf[pos] );	}

	case 's':
	    if  ( buf[pos+1] == 'S' )
		{ (*pPos) += 2; return ISO_germandbls;	}
	    else{ break;	}

	          case 'B': case 'C': case 'D': case 'E':
	case 'F': case 'G': case 'H': case 'I': case 'J':
	case 'K': case 'L': case 'M': case 'N':
	case 'P': case 'Q': case 'R': case 'S': case 'T':
	          case 'V': case 'W': case 'X': case 'Y':
	case 'Z':
	    (*pPos)++; return ISO_tolower( buf[pos] );
	}

    SSDEB((char *)buf,(char *)buf+pos);

    return -1;
    }

/************************************************************************/
/*  Read one affix rule.						*/
/************************************************************************/
static AffixRule *	affixNewRule(	int			type,
					int			shift,
					const unsigned char *	buf,
					AffixRule *		next	)
    {
    int		bufPos= 0;
    int		found;

    AffixRule *	rval;

    if  ( ! type )
	{ LDEB(type); return (AffixRule *)0;	}

    rval= (AffixRule *)malloc( sizeof(AffixRule) );
    if  ( ! rval )
	{ LDEB(rval); return rval;	}
    rval->arNext= next;
    rval->arWhatSide= type;
    rval->arShift= shift;

    if  ( buf[bufPos] == '.' )
	{
	rval->arMaskLength= 0; bufPos++;

	while( isspace(buf[bufPos]) )
	    { bufPos++;	}
	}
    else{
	if  ( buf[bufPos] == '>' )
	    { SSDEB((char *)buf,(char *)buf+bufPos); return (AffixRule *)0; }

	rval->arMaskLength= 0;
	for (;;)
	    {
	    if  ( rval->arMaskLength >= AFFamMAXLENGTH )
		{
		LLDEB(rval->arMaskLength,AFFamMAXLENGTH);
		return (AffixRule *)0;
		}

	    if  ( buf[bufPos] == '[' )
		{
		if  ( buf[bufPos+1] == '^' )
		    {
		    amFillMask( rval->arMasks+ rval->arMaskLength );
		    bufPos += 2;
		    while( buf[bufPos] != ']' )
			{
			found= affixChar( buf, &bufPos );
			if  ( found < 0 )
			    { SLDEB(buf,found); return (AffixRule *)0; }
			amClear( rval->arMasks+ rval->arMaskLength, found );
			}
		    bufPos++; /* skip ']' */
		    }
		else{
		    if  ( ! buf[bufPos+1] )
			{ SDEB(buf); return (AffixRule *)0; }

		    amEmptyMask( rval->arMasks+ rval->arMaskLength );
		    bufPos++;

		    while( buf[bufPos] != ']' )
			{
			found= affixChar( buf, &bufPos );
			if  ( found < 0 )
			    { SLDEB(buf,found); return (AffixRule *)0; }

			amSet( rval->arMasks+ rval->arMaskLength, found );
			}
		    bufPos++; /* skip ']' */
		    }
		}
	    else{
		if  ( ! buf[bufPos] )
		    { SDEB(buf); return (AffixRule *)0; }

		amEmptyMask( rval->arMasks+ rval->arMaskLength );
		found= affixChar( buf, &bufPos );
		if  ( found < 0 )
		    { SLDEB(buf,found); return (AffixRule *)0; }

		amSet( rval->arMasks+ rval->arMaskLength, found );
		}

	    rval->arMaskLength++;

	    while( isspace(buf[bufPos]) )
		{ bufPos++;	}

	    if  ( buf[bufPos] == '>' )
		{ break;	}
	    }
	}

    bufPos++; /* skip '>' */

    while( isspace(buf[bufPos]) )
	{ bufPos++;	}

    if  ( ! buf[bufPos] )
	{ SLDEB(buf,bufPos); return (AffixRule *)0; }

    if  ( buf[bufPos] == '-' )
	{
	bufPos++;
	rval->arFromLength= 0;
	for (;;)
	    {
	    if  ( rval->arFromLength >= AFFfrMAXLENGTH )
		{
		LLDEB(rval->arFromLength,AFFfrMAXLENGTH);
		return (AffixRule *)0;
		}

	    found= affixChar( buf, &bufPos );
	    if  ( found < 0 )
		{ SLDEB(buf,found); return (AffixRule *)0; }

	    rval->arToReplace[rval->arFromLength++]= found;
	    if  ( buf[bufPos] == ',' )
		{
		bufPos++;
		while( isspace(buf[bufPos]) )
		    { bufPos++;	}
		break;
		}
	    }
	}
    else{ rval->arFromLength= 0; }

    if  ( rval->arMaskLength < rval->arFromLength )
	{ LLDEB(rval->arMaskLength,rval->arFromLength); return (AffixRule *)0; }

    if  ( buf[bufPos] == '-' )
	{
	bufPos++;
	rval->arToLength= 0;
	}
    else{
	rval->arToLength= 0;
	while( buf[bufPos] && ! isspace( buf[bufPos] ) )
	    {
	    if  ( rval->arToLength >= AFFtoMAXLENGTH )
		{
		LLDEB(rval->arToLength,AFFtoMAXLENGTH);
		return (AffixRule *)0;
		}

	    found= affixChar( buf, &bufPos );
	    if  ( found < 0 )
		{ SLDEB(buf,found); return (AffixRule *)0; }
	    rval->arReplacement[rval->arToLength++]= found;
	    }
	}

    while( isspace(buf[bufPos]) )
	{ bufPos++;	}

    if  ( buf[bufPos] )
	{ SSDEB(buf,buf+bufPos); return (AffixRule *)0; }

    /*
    fprintf( stderr, "MaskLength=%d, arFromLength= %d, rval->arToLength=%d\n",
	    rval->arMaskLength, rval->arFromLength, rval->arToLength );
    */

    return rval;
    }

/************************************************************************/
/*  Read affix file.							*/
/************************************************************************/
int affixReadFile(	const char *	filename	)
    {
    FILE *	f= fopen( filename, "r" );
    int		pos;
    int		words;
    int		affixtype= 0;
    int		shift= 0;
    int		currentFlag= -1;
    int		l;
    int		line= 0;

    AffixRule *	newRule;

    if  ( ! f )
	{ SXDEB(filename,f); return -1;	}

    while( fgets( buf, BUL, f ) )
	{
	line++;

	if  ( ! strcmp( buf, "#up\n" ) )
	    { shift=  1; continue;	}
	if  ( ! strcmp( buf, "#down\n" ) )
	    { shift= -1; continue;	}

	l= 0;
	while( l < BUL && buf[l] != '#' && buf[l] != '\n' )
	    { l++;	}
	buf[l]= '\0';

	pos= 0;
	while( buf[pos] == ' ' || buf[pos] == '\t' )
	    { pos++;	}

	if  ( ! buf[pos] )
	    { continue;	}

	words= sscanf( buf, "%s%s", word1, word2 );

	if  ( words > 0 )
	    {
	    if  ( ! strcmp( word1, "defstringtype" )	)
		{ continue;	}
	    if  ( ! strcmp( word1, "compoundwords" )	)
		{ continue;	}
	    if  ( ! strcmp( word1, "allaffixes" )	)
		{ continue;	}
	    if  ( ! strcmp( word1, "wordchars" )	)
		{ continue;	}
	    if  ( ! strcmp( word1, "stringchar" )	)
		{ continue;	}
	    if  ( ! strcmp( word1, "altstringtype" )	)
		{ continue;	}
	    if  ( ! strcmp( word1, "altstringchar" )	)
		{ continue;	}

	    if  ( ! strcmp( word1, "prefixes" )	)
		{ affixtype= -1; continue;	}
	    if  ( ! strcmp( word1, "suffixes" )	)
		{ affixtype=  1; continue;	}
	    if  ( ! strcmp( word1, "flag" )	)
		{
		shift= 0;

		if  ( words > 1 )
		    {
		    if  ( word2[0] == '*' )
			{
			if  ( strlen( word2 ) == 3 && word2[2] == ':' )
			    { currentFlag= word2[1]; continue;	}
			}
		    else{
			if  ( strlen( word2 ) == 2 && word2[1] == ':' )
			    { currentFlag= word2[0]; continue;	}
			}
		    }
		}
	    }

	while( ! strchr( buf, '>' ) && l < BUL )
	    {
	    if  ( ! fgets( buf+ l, BUL- l, f ) )
		{ SLDEB(filename,line); return -1;	}

	    line++;

	    while( l < BUL && buf[l] != '#' && buf[l] != '\n' )
		{ l++;	}

	    if  ( buf[l] == '#' )
		{ buf[l]= '\0'; break; }
	    buf[l]= '\0';
	    }

	newRule= affixNewRule( affixtype, shift,
					buf+ pos, AFFRules[currentFlag] );
	if  ( ! newRule )
	    { SLDEB(filename,line); fclose( f ); return -1; }

	AFFRules[currentFlag]= newRule;
	}

    fclose( f ); return 0;
    }

int affixApplyRules(
    const unsigned char *	word,
    int				len,
    const unsigned char *	rules,
    void *			ind,
    int				(*fun)( void * ind,
					    const unsigned char * word ))
    {
    AffixRule *			ar;
    int				i, j;
    unsigned char *		target;
    const unsigned char *	tail;

    /*
    fprintf( stderr, "Apply rules: '%s'/'%s'\n", word, rules );
    */

    i= 0;
    while( rules[i] )
	{
	ar= AFFRules[rules[i]];

	while( ar )
	    {
	    if  ( ar->arMaskLength <= len )
		{
		if  ( ar->arWhatSide > 0 )
		    {
		    /*
		    printf("TRY %c: '%*.*s|%s' maskLen=%d\n",
			    rules[i],
			    len- ar->arMaskLength, len- ar->arMaskLength,
			    word,
			    word+ len- ar->arMaskLength,
			    ar->arMaskLength );
		    */

		    tail=word+ ( len- ar->arMaskLength );

		    for ( j= 0; j < ar->arMaskLength; j++ )
			{
			if  ( ! amIsSet( &(ar->arMasks[j]), tail[j] ) )
			    { break;	}
			}

		    if  ( j == ar->arMaskLength )
			{
			target= word1+ len- ar->arFromLength;
			memcpy( word1, word, len- ar->arFromLength );
			memcpy( target, ar->arReplacement, ar->arToLength );
			target[ar->arToLength]= '\0';

			/*
			target[0]= ISO_tolower( target[0] );
			*/

			if  ( ar->arShift > 0 )
			    { word1[0]= ISO_toupper( word1[0] ); }
			if  ( ar->arShift < 0 )
			    { word1[0]= ISO_tolower( word1[0] ); }

			/* printf("%c: %s\n", rules[i], word1 ); */

			if  ( (*fun)( ind, word1 ) < 0 )
			    { SDEB(word1); return -1;	}
			}
		    }

		if  ( ar->arWhatSide < 0 )
		    {
		    for ( j= 0; j < ar->arMaskLength; j++ )
			{
			if  ( ! amIsSet( &(ar->arMasks[j]), word[j] ) )
			    { break;	}
			}

		    if  ( j == ar->arMaskLength )
			{
			target= word1+ ar->arMaskLength;
			tail= word+ ar->arFromLength;
			memcpy( word1, ar->arReplacement, ar->arToLength );
			memcpy( target, tail, len- ar->arFromLength );
			target[len- ar->arFromLength]= '\0';

			if  ( ar->arShift > 0 )
			    { word1[0]= ISO_toupper( word1[0] ); }
			if  ( ar->arShift < 0 )
			    { word1[0]= ISO_tolower( word1[0] ); }

			/* printf("%c: %s\n", rules[i], word1 ); */

			if  ( (*fun)( ind, word1 ) < 0 )
			    { SDEB(word1); return -1;	}
			}
		    }
		}

	    ar= ar-> arNext;
	    }
	i++;
	}

    return 0;
    }
