;;; @ min-map.el - Minor mode keymap for Emacs
;;;
;;; $Id: min-map.el,v 1.12 1993/03/24 12:46:50 amanda Exp $

(provide 'min-map)

;;; @@ Copyright
;;;
;;; Copyright (C) 1993 Per Abrahamsen
;;; 
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; @@ Usage
;;;
;;; You can change a keymap in a minor mode, and automatically have it
;;; restored to the correct value when you leave the minor mode.  Note
;;; that the correct value might not be the original value, in the
;;; case where multiple minor modes redefine the same key and their
;;; invocation are not cleanly nested. 
;;;
;;; Use minor-set-key to set a key in the minor keymap for a given
;;; minor mode. 
;;;
;;; Use minor-call-shadow to call the key binding shadowed by a given
;;; minor mode.
;;;
;;; Use unbind-minor-mode to remove all key definitions for a given
;;; minor mode.

;;; @@ Minibuffer Usage
;;;
;;; You can set keys in any keymaps.
;;;
;;; Use minor-set-all to define a key in the local keymap and and all
;;; the minibuffer keymaps.
;;;
;;; When binding keys in the minibuffer, you must use minibuffer-value
;;; to get the value of a buffer local variable in the orifinal
;;; buffer.

;;; @@ Variable Usage
;;;
;;; You can define minor mode variables, that will automatically get
;;; restored to the correct value. when you leave the minor mode.
;;;
;;; Use minor-set-variable to set a variable locally in a minor mode. 
;;;
;;; Use minor-value-shadow to get the value of a shadowed variable. 
;;;
;;; Use unbind-minor-mode to restore all local variables for a given
;;; minor mode.

;;; @@ Description
;;;
;;; GNU Emacs minor modes may not change the keymap.  This is why modes
;;; like outline-mode and indented-text-mode are major modes.
;;; Minor modes could of cause just change the local key binding on entry,
;;; and then restore it on exit.  But this will fail, if some other
;;; minor mode changes the keybinding meanwhile.
;;;
;;; The idea behind minor-map is to keep a list of which minor mode has
;;; what binding for a particular key.  The list is sorted
;;; chronologically, with the most recently enabled minor mode first,
;;; and the major mode last.  The key will always be bound according to
;;; the head of the list, that is, the last enabled minor mode.  When a
;;; minor mode is disabled, it will be removed from the list, and if it
;;; was at the head of the list the key will be bound according to the
;;; next entry in the list. 
;;;
;;; When a minor mode is enabled, it should bind all its keys with
;;; minor-set-key, and, when disabled, unbound them with
;;; unbind-minor-mode. 
;;;
;;; By default all the functions operates on the local keymap, but all
;;; the functions take an optional argument, specifying which keymap
;;; to operate on.  minor-set-all will define a key for the local
;;; keymap, and for all minibuffer keymaps. 
;;;
;;; When rebinding keys in the minibuffer, you must use
;;; minibuffer-value to get the value of a buffer local variable in
;;; the original buffer.
;;;
;;; The minor mode keymaps are contained in a local variable
;;; minor-keymap-list.
;;;
;;; The minor maps might fail, if the major mode (or the user) change
;;; the local key bindings after the first call to minor-set-key.
;;;
;;; Minor maps should support shared keymaps for prefix keys.
;;;
;;; The support for variables was a quick hack, and the code ought to
;;; be restructured.
;;;
;;; It should be possible to document minor modes.

;;; @@ Minor Keymap List
;;;
;;; We need a local variable to hold the keybindings for all the
;;; keymaps.  

(defvar minor-keymap-list nil
  "Association list of keymaps with keys remapped by minor modes.

The syntax of the list is:
	(KEYMAP . (KEY . (MODE . COMMAND) ...) ...) ...

For each KEYMAP, an association list of keys is maintained.  
For each KEY, an list of modes is maintained.
For each MODE, the keys definition in the keymap for that mode is stored.

For each KEYMAP, every KEY represents a command that has been bound by
a minor mode at some point.  If a key sequence have no entry, it means
that the KEY is bound to the default value in the major mode map.

NAME identify a minor mode, that have bound KEY to COMMAND in KEYMAP.
The first MODE in the list represent the newest, and currently active,
binding.  The MODE at the tail of the list is always the major mode,
and the COMMAND is the default binding in the major mode. The MODE
list will always contain the entry for the major mode.

This variable should not be manipulated directly, use instead the
functions `minor-set-key' and  `unbind-minor-mode'.")

  (make-variable-buffer-local 'minor-keymap-list)

;;; @@ Keymaps

(defun minor-set-key (key command name &optional keymap)
  "Bind KEY to COMMAND in the minor mode NAME.
NAME is a symbol that must identify the minor mode.

If the optional fourth argument KEYMAP is not nil, it will be used
intstead of the current local map."

  ;; Initialize local keymaps if needed.
  (minor-keymap-initialize)

  ;; If no keymap given, assume local map. 
  (if (not keymap)
      (setq keymap (current-local-map)))

  ;; Look the keymap up in the list.
  (let ((map-entry (assoc keymap minor-keymap-list)))

    ;; Create entry for the keymap if it did not exist before
    (if map-entry
	()
      (setq map-entry (cons keymap nil))
      (setq minor-keymap-list (cons map-entry minor-keymap-list)))
      
    ;; Lookup the key in the map.
    (let ((key-entry (assoc key (cdr map-entry))))
      ;; Does it already exist?
      (if key-entry
	  ()
	;; Otherwise create it.
	(let ((old (cond ((keymapp keymap)
			  (let ((lookup (lookup-key keymap key)))
			    (if (or (null lookup) (numberp lookup))
				(lookup-key (current-global-map) key)
			      lookup)))
			 ((eq keymap 'variable)
			  (and (boundp key)
			       (symbol-value key)))
			 (t
			  (error "Unknown map type '%s'" keymap)))))
	  (setq key-entry
		(list key (cons major-mode old))))
	(setcdr map-entry (cons key-entry (cdr map-entry))))
      ;; Put this binding in front of the list.
      (if (assoc name (cdr key-entry))
	  ;; First check that it not already is in the list.
	  (error "Minor key `%s' already defined for mode `%s'." key name)
	(setcdr key-entry (cons (cons name command) (cdr key-entry)))))
    ;; Make the binding active.
    (cond ((keymapp keymap)
	   (define-key keymap key command))
	  ((eq keymap 'variable)
	   (set key command))
	  (t
	   (error "Unknown map type '%s'" keymap)))))

(defun minor-call-shadow (mode key &optional keymap)
  "Call the function shadowed by MODE for KEY interactively.

If the optional third argument KEYMAP is not nil, it will be used
intstead of the current local map."

  ;; If no keymap given, assume local map. 
  (if (not keymap)
      (setq keymap (current-local-map)))

  (let (map-entry key-entry mode-entry mode-iterator)
    (if (and (setq map-entry (assoc keymap
				    (minibuffer-value 'minor-keymap-list)))
	     (setq key-entry (assoc key (cdr map-entry)))
	     ;; Finding the mode entry is a bit tricky, since we want
	     ;; the mode entry _after_ the entry containing MODE.
	     (setq mode-entry
		   (progn
		     (setq mode-iterator (cdr key-entry))
		     (while (and (cdr mode-iterator)
				 (not (equal (car (car mode-iterator)) mode)))
		       (setq mode-iterator (cdr mode-iterator)))
		     (car-safe (cdr-safe mode-iterator)))))
	;; Found
	(cond ((keymapp keymap)
	       (cond ((commandp (cdr mode-entry))
		      ;; Bound in the keymap.  Execute.
		      (command-execute (cdr mode-entry)))
		     ((null (cdr mode-entry))
		      (beep))
		     ((numberp (cdr mode-entry))
		      ;; Old inline prefix.  Cheat and insert verbatim.
		      (insert key))
		     ((keymapp (cdr mode-entry))
		      (let ((index (char-to-string (read-char))))
			(while (keymapp (lookup-key (cdr mode-entry) index))
			  (setq index (concat index
					      (char-to-string (read-char)))))
			(cond ((commandp (lookup-key (cdr mode-entry) index))
			       (command-execute (lookup-key (cdr mode-entry)
							    index)))
			      ((commandp (lookup-key (current-global-map)
						     (concat key index)))
			       (command-execute
				(lookup-key (current-global-map)
					    (concat key index))))
			      ;; Here we should test for keymapp and
			      ;; search further in the global map.
			      ;; * some other time *
			      (t
			       (beep)))))
		     (t
		      ;; Anything else. What could that be?
		      (beep))))
	      ((eq keymap 'variable)
	       ;; A variable
	       (cdr mode-entry))
	      (t
	       (error "Unknown map type '%s'" keymap)))
      ;; Not found.  
      (cond ((keymapp keymap)
	     ;; When a keymap is not found, there might be several
	     ;; reasons.   One of the uglier is that the current key
	     ;; originally was a prefix key.  This should be handled
	     ;; somehow by minor-set-key, but currently we just cheat and
	     ;; insert the string verbatim.
	     (insert key))
	    ((eq keymap 'variable)
	     ;; All variables default to nil.
	     nil)
	    (t
	     (error "Unknown map type '%s'" keymap))))))

(defun unbind-minor-mode (name &optional keymap)
  "Remove all references to minor mode NAME from the all keymaps.

If the optional second argument KEYMAP is not nil, only that keymap
will be affected.
Key bindings will be updated to reflect the new content of the minor
keymap."

  (let ((map-entry (if keymap
		       (list (assoc keymap minor-keymap-list))
		     minor-keymap-list)))
    (while map-entry
      (let ((keymap (car (car map-entry)))
	    (key-entry (cdr (car map-entry))))
	(while key-entry
	  (let* ((entry (car key-entry))
		 (key (car entry))
		 (current (cdr entry))
		 (previous entry))
	    ;; Reset binding to the one shadowed.
	    (if (eq name (car (car current)))
		(cond ((keymapp keymap)
		       (define-key keymap key (cdr (car (cdr current)))))
		      ((eq keymap 'variable)
		       (set key (cdr (car (cdr current)))))
		      (t
		       (error "Unknown map type '%s'" keymap))))
	    ;; Remove this mode from the keys mode list.
	    (while current
	      (if (eq name (car (car current)))
		  (setcdr previous (cdr current))
		(setq previous current))
	      (setq current (cdr current))))
	  (setq key-entry (cdr key-entry))))
      (setq map-entry (cdr map-entry)))))

;;; @@ Variables

(defun minor-set-variable (variable value name)
  "Set VARIABLE to VALUE in the minor mode NAME.
NAME is a symbol that must identify the minor mode.

The variable will get its value restored correctly, when the minor
mode is left."
  (make-variable-buffer-local variable)
  (minor-set-key variable value name 'variable))

(defun minor-value-shadow (mode variable)
  "Find the value shadowed by MODE of VARIABLE."
  (minor-call-shadow mode variable 'variable))

;;; @@ Initialization
;;;
;;; We need to make sure the keymaps we change are really buffer
;;; local, since we don't want minor modes to affect more than one
;;; buffer. 
;;; 
;;; This is needed because - despite its name - the local keymap is
;;; often just a pointer to keymap shared between all buffers in the
;;; same major mode. 
;;;
;;; We do this by copying all the standard local keymaps first time
;;; the user tries to modify any keymap.

(defvar minor-keymap-initialized nil
  "TRUE, if we have made a copy of the local keymap.")

  (make-variable-buffer-local 'minor-keymap-initialized)

(defun minor-keymap-initialize ()
  "Initialize all buffer local keymaps."
  
  (if minor-keymap-initialized
      ()
    (setq minor-keymap-initialized t)

    (use-local-map (if (current-local-map)
		       (copy-keymap (current-local-map))
		     (make-sparse-keymap)))
    
    (make-local-variable 'minibuffer-local-completion-map)
    (setq minibuffer-local-completion-map
	  (copy-keymap minibuffer-local-completion-map))
    (make-local-variable 'minibuffer-local-map)
    (setq minibuffer-local-map
	  (copy-keymap minibuffer-local-map))
    (make-local-variable 'minibuffer-local-must-match-map)
    (setq minibuffer-local-must-match-map
	  (copy-keymap minibuffer-local-must-match-map))))

;;; @@ Minibuffer
;;; 
;;; When executing a function bound to a key in the minibuffer, we can
;;; not access the local variables of the normal buffer directly.
;;;
;;; This hack assumes that the normal buffer is displayed in the first
;;; other window, which is often the case.
;;;
;;; If someone suggest a more reliable way to get the buffer from
;;; which the minibuffer was activated, I'm all Perot.

(defun minibuffer-value (symbol)
  "Return the value of symbol.

If this is evaluated in the minibuffer, use value from other window."
  (if (eq (minibuffer-window) (selected-window))
      (save-window-excursion
	(other-window 1)
	(symbol-value symbol))
    (symbol-value symbol)))

(defun minor-set-all (key command name)
  "Bind KEY to COMMAND in the minor mode NAME.
NAME is a symbol that must identify the minor mode.

The key will be bound for the local keymap and for all minnibuffer
keymaps."
  (minor-set-key key command name minibuffer-local-completion-map)
  (minor-set-key key command name minibuffer-local-map)
  (minor-set-key key command name minibuffer-local-must-match-map)
  (minor-set-key key command name))

;;; @@ Emacs

;;; Local Variables:
;;; mode: emacs-lisp
;;; mode: outline-minor
;;; outline-regexp: ";;; @+\\|(......"
;;; End:
