;;; @ tex-auto.el - Automatically generate style information.
;;; 
;;; $Id: tex-auto.el,v 5.5 1993/03/19 01:12:42 amanda Exp $

(provide 'tex-auto)
(require 'tex-init)

;;; @@ Copyright
;;;
;;; Copyright (C) 1993 Per Abrahamsen 
;;; 
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; @@ Buffer Local Information

(defun TeX-auto-find-files ()
  "Find TeX information in the current buffer."
  (TeX-auto-parse)
  
  (if TeX-auto-symbol
      (apply 'TeX-add-symbols TeX-auto-symbol))

  (if TeX-auto-environment
      (apply 'LaTeX-add-environments TeX-auto-environment))
  (if TeX-auto-label
      (apply 'LaTeX-add-labels TeX-auto-label))
  (if TeX-auto-bibliography
      (apply 'LaTeX-add-bibliographies TeX-auto-bibliography))
  (if TeX-auto-style
      (setq LaTeX-style (nth 0 TeX-auto-style)
            LaTeX-options (nth 1 TeX-auto-style)))

  TeX-auto-file)

;;; @@ Local Auto Directory

(defvar TeX-auto-update nil
  "Automatically update the lisp information for this TeX file.")

 (make-variable-buffer-local 'TeX-auto-update)

(defun TeX-auto-write-hook ()
  "Save all relevant TeX information from the current buffer."
  (if (not TeX-auto-update)
      ()
    (untabify (point-min) (point-max))
    (if TeX-auto-local
        (let ((file (concat TeX-auto-local
                            (if (string-match "/$" TeX-auto-local) "" "/")
                            (TeX-strip-extension nil TeX-all-extensions t)
                            ".el")))
          (if (file-writable-p file)
              (save-excursion
                (TeX-update-style)
                (TeX-auto-store file))))))
  ;; Continue with the other write file hooks.
  nil)

;;; @@ Private Auto Directory

(defvar TeX-macro-default (car TeX-macro-private)
  "*Default directory to search for TeX macros.")

(defvar TeX-auto-default TeX-auto-private
  "*Default directory to place automatically generated TeX information.")

(defun TeX-auto-generate (tex auto)
  "Generate style file for TEX and store it in AUTO.  
If TEX is a directory, generate style files for all files in the directory."
  (interactive (list (setq TeX-macro-default
                           (expand-file-name (read-file-name
                                              "TeX file or directory: "
                                              TeX-macro-default
                                              "." 'confirm)))
                     (setq TeX-auto-default
                           (expand-file-name (read-file-name
                                              "AUTO lisp directory: "
                                              TeX-auto-default
                                              "." 'confirm)))))
  (cond ((not (file-readable-p tex)))
	((string-match TeX-ignore-file tex))
        ((file-directory-p tex)
         (let ((files (directory-files tex))
               (default-directory (concat (if (string-match "^/" tex)
                                              ""
                                            default-directory)
                                          (if (string-match "/$" tex)
                                              tex
                                            (concat tex "/")))))
           (mapcar (function (lambda (file) (TeX-auto-generate file auto)))
                   files)))
        ((TeX-match-extension tex)
         (save-excursion
           (set-buffer (find-file-noselect tex))
           (message "Parsing %s..." tex)
           (TeX-auto-store (concat auto
                                   (if (string-match "/$" auto) "" "/")
                                   (TeX-strip-extension tex nil t)
                                   ".el"))
           (kill-buffer (current-buffer))
           (message "Parsing %s... done" tex)))))

;;; @@ Global Auto Directory

(defun TeX-auto-generate-global ()
  "Create global auto directory for global TeX macro definitions."
  (interactive)
  (mapcar (function (lambda (macro) (TeX-auto-generate macro TeX-auto-global)))
          TeX-macro-global))

;;; @@ Store File

(defun TeX-auto-store (file)
  "Extract information for auc tex from current buffer and store it in FILE."
  (TeX-auto-parse)
  
  (if (or TeX-auto-symbol
          TeX-auto-environment
          TeX-auto-file
          TeX-auto-label
          TeX-auto-bibliography)
      (let ((style (TeX-strip-extension nil TeX-all-extensions t)))
        (TeX-unload-style style)
        (set-buffer (find-file-noselect file))
        (erase-buffer)
        (insert "(TeX-add-style-hook \"" style "\"\n"
                " (function\n"
                "  (lambda ()")
        (TeX-auto-insert "TeX-add-symbols" TeX-auto-symbol)
        (TeX-auto-insert "LaTeX-add-environments" TeX-auto-environment)
        (TeX-auto-insert "LaTeX-add-labels" TeX-auto-label)
        (TeX-auto-insert "LaTeX-add-bibliographies" TeX-auto-bibliography)
        (TeX-auto-insert "TeX-run-style-hooks" TeX-auto-file)
        (insert ")))")
        (if (> (current-column) fill-column)
            (do-auto-fill))
        (insert "\n\n")
        (save-buffer)
        (kill-buffer (current-buffer)))))

(defun TeX-auto-insert (name list)
  "Insert hook NAME in current buffer called with LIST."
  (if (null list)
      ()
    (insert "\n    (" name)
    (while list
      (if (stringp (car list))
          (insert " " (prin1-to-string (car list)))
        (newline-and-indent)
        (insert "'" (prin1-to-string (car list))))
      (if (> (current-column) fill-column)
          (do-auto-fill))
      (setq list (cdr list)))
    (insert ")")
    (if (> (current-column) fill-column)
        (do-auto-fill))))

;;; @@ Data

(defvar TeX-auto-ignore
  '("csname" "filedate" "fileversion" "docdate" "next" "labelitemi"
    "labelitemii" "labelitemiii" "labelitemiv" "labelitemv"
    "labelenumi" "labelenumii" "labelenumiii" "labelenumiv"
    "labelenumv" "theenumi" "theenumii" "theenumiii" "theenumiv"
    "theenumv" "document" "par" "do" "expandafter")
  "List of symbols to ignore when scanning a TeX style file.")

(defvar TeX-auto-regexp-list
  '(("\\\\def\\\\\\([a-zA-Z]+\\)[^a-zA-Z@]" 1 TeX-auto-symbol-hook)
    ("\\\\let\\\\\\([a-zA-Z]+\\)[^a-zA-Z@]" 1 TeX-auto-symbol-hook)
    ("\\\\font\\\\\\([a-zA-Z]+\\)[^a-zA-Z@]" 1 TeX-auto-symbol)
    ("\\\\chardef\\\\\\([a-zA-Z]+\\)[^a-zA-Z@]" 1 TeX-auto-symbol)
    ("\\\\new\\(count|dimen|muskip|skip\\)\\\\\\([a-z]+\\)[^a-zA-Z@]"
     2 TeX-auto-symbol)
    ("\\\\newcommand{?\\\\\\([a-zA-Z]+\\)}?\\[\\([0-9]+\\)\\]"
     (1 2) TeX-auto-arguments)
    ("\\\\newcommand{?\\\\\\([a-zA-Z]+\\)}?" 1 TeX-auto-symbol)
    ("\\\\newfont{?\\\\\\([a-zA-Z]+\\)}?" 1 TeX-auto-symbol)
    ("\\\\typein\\[\\\\\\([a-zA-Z]+\\)\\]" 1 TeX-auto-symbol)
    ("\\\\newenvironment{\\([a-zA-Z]+\\)}" 1 TeX-auto-environment)
    ("\\\\newtheorem{\\([a-zA-Z]+\\)}" 1 TeX-auto-environment)
    ("\\\\input{\\([^#\\\\\\.\n]+\\)\\(\\.sty\\)?}" 1 TeX-auto-file)
    ("\\\\input +\\([^#\\\\\\.\n]+\\)\\(\\.sty\\)?" 1 TeX-auto-file)
    ("\\\\mathchardef\\\\\\([a-zA-Z]+\\)[^a-zA-Z@]" 1 TeX-auto-symbol)
    ("\\\\label{\\([]a-zA-Z0-9.:;,?!`'()[---/*@]+\\)}" 1 TeX-auto-label)
    ("\\\\bibliography{\\([^#\\\\\\.\n]+\\)}" 1 TeX-auto-bibliography)
    ("\\\\documentstyle\\[\\([^#\\\\\\.\n]+\\)\\]{\\([^#\\\\\\.\n]+\\)}"
     (2 1) TeX-auto-style)
    ("\\\\documentstyle{\\([^#\\\\\\.\n]+\\)}" (1) TeX-auto-style))
  "List of regular expression matching TeX macro definitions.

The list has the following format: ((REGEXP MATCH TABLE) ...), that
is, each entry is a list with three elements.

REGEXP.  Regular expression matching the macro we want to parse.

MATCH. A number or list of numbers, each representing one
parenthesized subexpression macthed by REGEXP.

TABLE.  The symbol table to store the data.  This can be function, in
which case the function is called with the argument MATCH.  Use
TeX-match-buffer to get match data.  If it is not a function, it is
presumed to be the name of a variable containing a list of match data.
The macthed data (a string if MATCH is a number, a list of strings if
MATCH is a list of numbers) is put in front of the table.")

;;; @@ Variables

(defvar TeX-auto-symbol nil
  "Local variable used when parsing TeX code.")

(defvar TeX-auto-arguments nil
  "Local variable used when parsing TeX code.")

(defvar TeX-auto-environment nil
  "Local variable used when parsing TeX code.")

(defvar TeX-auto-file nil
  "Local variable used when parsing TeX code.")

(defvar TeX-auto-label nil
  "Local variable used when parsing TeX code.")

(defvar TeX-auto-bibliography nil
  "Local variable used when parsing TeX code.")

(defvar TeX-auto-style nil
"Local variable used when parsing TeX code.")

(defvar TeX-auto-end-symbol nil
  "Local variable used when parsing TeX code.")

;;; @@ Parse File

(defvar TeX-auto-prepare-hooks nil
  "List of hooks to be called before parsing a TeX file.")

(defvar TeX-auto-cleanup-hooks nil
  "List of hooks to be called after partsing a TeX file.")

(defun TeX-auto-parse ()
  "Parse TeX information in current buffer.

Call the functions in TeX-auto-prepare-hooks before parsing, and the
functions in TeX-auto-cleanup-hooks after parsing."

  (let ((case-fold-search nil))

    ;; Prepare 
    (setq TeX-auto-symbol nil
	  TeX-auto-arguments nil
	  TeX-auto-environment nil
	  TeX-auto-file nil
	  TeX-auto-label nil
	  TeX-auto-bibliography nil
	  TeX-auto-style nil
	  TeX-auto-end-symbol nil)
    (mapcar 'funcall TeX-auto-prepare-hooks)
    
    ;; Parse
    (save-excursion
      (goto-char (point-max))
      ;; Extract the information.
      (let ((regexp (mapconcat (function (lambda (entry)
					   (concat "^[^%\n]*" (nth 0 entry))))
			       TeX-auto-regexp-list "\\|")))
	(while (re-search-backward regexp nil t)
	  (let* ((entry (TeX-member nil TeX-auto-regexp-list
				    (function (lambda (a b)
						(looking-at
						 (concat "^[^%\n]*"
							 (nth 0 b)))))))
		 (symbol (nth 2 entry))
		 (match (nth 1 entry)))
	    (if (fboundp symbol)
		(funcall symbol match)
	      (set symbol (cons (if (listp match)
				    (mapcar 'TeX-match-buffer match)
				  (TeX-match-buffer match))
				(symbol-value symbol))))))))
    
    ;; Cleanup ignored symbols.
    
    ;; NOTE: This is O(N M) where it could be O(N log N + M log M) if we 
    ;; sorted the lists first.
    (while (TeX-member (car TeX-auto-symbol) TeX-auto-ignore 'string-equal)
      (setq TeX-auto-symbol (cdr TeX-auto-symbol)))
    (let ((list TeX-auto-symbol))
      (while (and list (cdr list))
	(if (TeX-member (car (cdr list)) TeX-auto-ignore 'string-equal)
	    (setcdr list (cdr (cdr list)))
	  (setq list (cdr list)))))
    
    ;; Cleanup BibTeX files
    (setq TeX-auto-bibliography
	  (apply 'append (mapcar (function (lambda (arg)
					     (TeX-split-string "," arg)))
				 TeX-auto-bibliography)))
    
    ;; Cleanup document styles
    (if (null TeX-auto-style)
	()
      ;; We only use the first document style found.
      (setq TeX-auto-style (car TeX-auto-style))
      ;; We add the style file
      (setq TeX-auto-file (cons (car TeX-auto-style) TeX-auto-file))
      ;; And the options
      (if (nth 1 TeX-auto-style)
	  (let ((options (TeX-split-string "," (nth 1 TeX-auto-style))))
	    (setcdr TeX-auto-style options)
	    (setq TeX-auto-file (append options TeX-auto-file))))
      ;; And a special "art10" style file combining style and size.
      (setq TeX-auto-file
	    (cons (concat (cond ((string-equal "article" (car TeX-auto-style))
				 "art")
				((string-equal "book" (car TeX-auto-style))
				 "bk")
				((string-equal "report" (car TeX-auto-style))
				 "rep")
				(t
				 (nth 0 TeX-auto-style)))
			  (cond ((TeX-member "11pt" TeX-auto-style
					     'string-equal)
				 "11")
				((TeX-member "12pt" TeX-auto-style
					     'string-equal)
				 "12")
				(t
				 "10")))
		  TeX-auto-file)))
    
    ;; Cleanup optional arguments
    (mapcar (function (lambda (entry)
			(setq TeX-auto-symbol
			      (cons (list (nth 0 entry)
					  (string-to-int (nth 1 entry)))
				    TeX-auto-symbol))))
	    TeX-auto-arguments)
    
    ;; Cleanup use of def to add environments
    ;; NOTE: This use a O(N^2) algorithm, while a O(N log N) algorthm is
    ;; possible.
    (mapcar (function (lambda (symbol)
			(if (not (TeX-member symbol TeX-auto-symbol 'equal))
			    ;; No matching symbol, insert in list
			    (setq TeX-auto-symbol
				  (cons (concat "end" symbol) TeX-auto-symbol))
			  ;; Matching symbol found, remove from list
			  (if (equal (car TeX-auto-symbol) symbol)
			      ;; Is it the first symbol?
			      (setq TeX-auto-symbol (cdr TeX-auto-symbol))
			    ;; Nope!  Travel the list
			    (let ((list TeX-auto-symbol))
			      (while (consp (cdr list))
				;; Until we find it.
				(if (equal (car (cdr list)) symbol)
				    ;; Then remove it.
				    (setcdr list (cdr (cdr list))))
				(setq list (cdr list)))))
			  ;; and add the symbol as an environment.
			  (setq TeX-auto-environment
				(cons symbol TeX-auto-environment)))))
	    TeX-auto-end-symbol)
    
    (mapcar 'funcall TeX-auto-cleanup-hooks)))

(defun TeX-auto-symbol-hook (symbol)
  "Add MATCH to TeX-auto-symbol.
Check for potential LaTeX environments."
  (let ((symbol (if (listp match)
                    (mapcar 'TeX-match-buffer match)
                  (TeX-match-buffer match))))
    (if (and (stringp symbol)
             (string-match "^end\\(.+\\)$" symbol))
        (setq TeX-auto-end-symbol
              (cons (substring symbol (match-beginning 1) (match-end 1))
                    TeX-auto-end-symbol))
      (setq TeX-auto-symbol (cons symbol TeX-auto-symbol)))))

;;; @@ Emacs

(run-hooks 'TeX-after-tex-auto-hook)

;;; Local Variables:
;;; mode: emacs-lisp
;;; mode: outline-minor
;;; outline-regexp: ";;; @+\\|(......"
;;; End:


