;;; @ tex-math.el - Smart abbrevation for common math expressions.
;;; 
;;; $Id: tex-math.el,v 5.22 1993/03/18 08:45:53 amanda Exp $

(require 'min-map)

;;; @@ Copyright
;;;
;;; Copyright (C) 1991 Kresten Krab Thorup (krab@iesd.auc.dk).
;;; Copyright (C) 1993 Per Abrahamsen

;;; TO DO LIST
;;;
;;; Make the documentation be updated on changes to the list
;;;
;;; 
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; @@ Customization

(defvar LaTeX-math-abbrev-prefix "`"
  "*Prefix key for use in LaTeX-math-mode.")

;;; @@ The Mode

(defvar TeX-math-keymap (make-keymap)
  "Keymap for math abbreviations.")

(defvar TeX-math-mode nil
  "Is LaTeX-math-mode on or off? non nil means on.")

(if (not (assoc 'TeX-math-mode minor-mode-alist))
    (progn
      (setq minor-mode-alist (cons '(TeX-math-mode " Math") minor-mode-alist))
      (make-variable-buffer-local 'TeX-math-mode)))

(defun LaTeX-math-mode ()
  "A minor mode with easy acces to TeX math macros. 
Further documentation will be present when the mode has been set."
  (interactive)
  (if TeX-math-mode
      (unbind-minor-mode 'TeX-math-keymap)
    (minor-set-key LaTeX-math-abbrev-prefix
		   'LaTeX-abbrev-math
		   'TeX-math-keymap))
  (setq TeX-math-mode (not TeX-math-mode))
  (set-docstring
   'LaTeX-math-mode
   (concat
    "A minor mode, in which " LaTeX-math-abbrev-prefix " is"
    " bound to a list of standard math symbols.\n"
    "With prefix (C-u) the symbol is inserted in single dollar signs.\n"
    "The prefix char may be customized by changing the value of the\n"
    "value of the string variable LaTeX-math-abbrev-prefix.\n\n"
    "Current keybindings are:\n\n"
    LaTeX-math-abbrev-prefix " ?   Help (this list)\n"

    (let ((list LaTeX-mathlist)
	  (endlist ""))
      (while (cdr list)
	(progn
	  (setq endlist
		(concat
		 endlist
		 LaTeX-math-abbrev-prefix
		 " "
		 (char-to-string (car (car list)))
		 "   \\"
		 (cdr (car list))
		 "\n"))
	  (setq list (cdr list))))
      endlist)))
  ;; force modeline update soon
  (set-buffer-modified-p (buffer-modified-p)))

;;; @@ Utility

(defun set-docstring (symbol string)
  "Set the documentation string for symbol SYMBOL to STRING."
  (let* ((symfunc (symbol-function symbol))
	 (doc-cdr (nthcdr 2 symfunc)))
    (condition-case nil 
	  (setcar doc-cdr string)
	(error
	 (setq symfunc (copy-sequence (symbol-function
				       symbol))
	       doc-cdr (nthcdr 2 symfunc))
	 (setcar doc-cdr string)
	 (fset symbol symfunc)))))

;;; @@ Data

(defvar LaTeX-mathlist '(( ?a . "alpha" )
		      ( ?b . "beta" )
		      ( ?d . "delta" )
		      ( ?e . "epsilon" )
		      ( ?f . "phi" )
		      ( ?g . "gamma" )
		      ( ?h . "eta" )
		      ( ?k . "kappa" )
		      ( ?l . "lambda" )
		      ( ?m . "mu" )
		      ( ?n . "nu" )
		      ( ?o . "omega" )
		      ( ?p . "pi" )
		      ( ?q . "theta" )
		      ( ?r . "rho" )
		      ( ?s . "sigma" )
		      ( ?t . "tau" )
		      ( ?u . "upsilon" )
		      ( ?x . "chi" )
		      ( ?y . "psi" )
		      ( ?z . "zeta" )
		      ( ?D . "Delta" )
		      ( ?G . "Gamma" )
		      ( ?Q . "Theta" )
		      ( ?L . "Lambda" )
		      ( ?Y . "Psi" )
		      ( ?P . "Pi" )
		      ( ?S . "Sigma" )
		      ( ?U . "Upsilon" )
		      ( ?V . "Phi" )
		      ( ?O . "Omega" )
		      ( ?\C-f . "rightarrow" )
		      ( ?\C-b . "leftarrow" )
		      ( ?\C-p . "uparrow" )
		      ( ?\C-n . "downarrow" )
		      ( ?< . "leq" )
		      ( ?> . "geq" )
		      ( ?~ . "tilde" )
		      ( ?I . "infty" )
		      ( ?A . "forall" )
		      ( ?E . "exists" )
		      ( ?! . "not" )
		      ( ?i . "in" )
		      ( ?* . "times" )
		      ( ?. . "cdot" )
		      ( ?{ . "subset" )
		      ( ?} . "supset" )
		      ( ?[ . "subseteq" )
		      ( ?] . "supseteq" )
		      ( ?\\ . "backslash" )
		      ( ?/ . "setminus" )
		      ( ?+ . "cup" )
		      ( ?- . "cap" )
		      ( ?( . "langle")
		      ( ?) . "rangle")
		      ( ?\C-e . "exp")
		      ( ?\C-s . "sin")
		      ( ?\C-c . "cos")  
		      ( ?\C-^ . "sup" )
		      ( ?\C-_ . "inf" )
		      ( ?\C-d . "det" )
		      ( ?\C-l . "lim" )
		      ( ?\C-t . "tan" )
		      ( ?^ . "hat")
		      ( ?v . "vee")
		      ( ?0 . "emptyset" ))
  "Alist of key-symbol mappings to use in LaTeX-math-mode.
The user may define new mapping by redefining this variable.

To expand the current mapping by binding \"\\alpha\" to `-a you may insert the followin in your .emacs:

   (setq TeX-after-tex-math-hook (function (lambda ()
                                            (setq LaTeX-mathlist
                                                  (cons '( ?a . \"alpha\")
			                           LaTeX-mathlist)))))")

;;; @@ Interactive Commands

(defun LaTeX-abbrev-math (p char)
  "Smart bindings of math symbols. See LaTeX-math-mode for documentation."
  (interactive "p\nc")
  (cond ((= (string-to-char LaTeX-math-abbrev-prefix) char)
	 (insert-string LaTeX-math-abbrev-prefix))
	((= ?? char)
	 (describe-function 'LaTeX-math-mode))
	((assoc char LaTeX-mathlist)
	 (insert (concat (if (= p 1 ) "" "$")
			 "\\" (cdr (assoc char LaTeX-mathlist))
			 (if (= p 1 ) "" "$"))))
	(t
	 (error "no such LaTeX-math-abbrev defined"))))

;;; @@ Emacs

(run-hooks 'TeX-after-tex-math-hook)

;;; Local Variables:
;;; mode: emacs-lisp
;;; mode: outline-minor
;;; outline-regexp: ";;; @+\\|(......"
;;; End:
