/*
 * rtfwc - read rtf input, write word count (actually, char, word
 * 	and paragraph counts).
 *
 * The wrapper must provide an Output() function for sending the output
 * somewhere.  The UNIX version just sends it to stdout, but the Macintosh
 * version routes the output into the log window.
 *
 * This installs callbacks for the ascii and control token classes.
 * The control class is necessary so that special characters such as
 * \par, \tab, \sect, etc.  can be counted.
 *
 * Counts paragraphs instead of lines, since the concept of "line"
 * is relatively meaningless.
 *
 * It's problematic how to count text in headers and footers, and
 * what to do about tables.
 *
 * 04 Feb 91	Paul DuBois	dubois@primate.wisc.edu
 *
 * 04 Feb 91 V1.0
 * - Created.
 * 27 Feb 91 V1.01
 * - Updated for distribution 1.05.
 * 16 Mar 91 V1.02
 * - Updated for distribution 1.06.  Multiple files allowed.
 * 14 Jul 93 V1.03
 * - Updated to compile under THINK C 6.0.
 * 18 Mar 94
 * - Updated for distribution 1.10.
 */

# include	<stdio.h>

# include	"rtf.h"
# include	"rtfwc.h"


static void	TextClass ();
static void	ControlClass ();

static int	nFiles = 0;
static long	chars, tchars = 0;
static long	words, twords = 0;
static long	paras, tparas = 0;
static long	wchars = 0;	/* chars in current word */


/*
 * Done reading all files.  Print stats totals.
 */

void
Summarize ()
{
char	buf[rtfBufSiz];

	if (nFiles > 1)	/* multiple files, print totals */
	{
		sprintf (buf, "%ld chars\t%ld words\t%ld paragraphs\ttotal\n",
					tchars, twords, tparas);
		Output (buf);
	}
}


/*
 * Set up to begin counting a new file.  The input file name is remembered
 * for later with the file stats are printed.  The output file name is
 * ignored because all output goes to Output().
 */

int
BeginFile ()
{
	++nFiles;
	chars = words = paras = wchars = 0;

	/* install counting hooks and process the input stream */

	RTFSetClassCallback (rtfText, TextClass);
	RTFSetClassCallback (rtfControl, ControlClass);

	return (1);
}


/*
 * Done processing file.  Print stats.
 */

void
EndFile ()
{
char	buf[rtfBufSiz];

	sprintf (buf, "%ld chars\t%ld words\t%ld paragraphs",
						chars, words, paras);
	Output (buf);
	if (RTFGetInputName () != (char *) NULL)
	{
		sprintf (buf, "\t%s", RTFGetInputName ());
		Output (buf);
	}
	Output ("\n");
	tchars += chars;
	twords += words;
	tparas += paras;
}


static void
TextClass ()
{
	++chars;
	if (rtfMajor != ' ')
		++wchars;
	else if (wchars > 0)
		++words;
}


static void
ControlClass ()
{
	if (rtfMajor != rtfSpecialChar)
		return;
	switch (rtfMinor)
	{
	case rtfPage:
	case rtfSect:
	case rtfLine:
	case rtfPar:
		++paras;
		++chars;
		if (wchars > 0)
		{
			++words;
			wchars = 0;
		}
		break;
	case rtfNoBrkSpace:
	case rtfTab:
		++chars;
		if (wchars > 0)
		{
			++words;
			wchars = 0;
		}
		break;
	case rtfNoBrkHyphen:
	case rtfBullet:
	case rtfEmDash:
	case rtfEnDash:
	case rtfLQuote:
	case rtfRQuote:
	case rtfLDblQuote:
	case rtfRDblQuote:
		++chars;
		++wchars;
		break;
	}
}
